const WebSocketServer = require('websocket').server;
const { URLSearchParams } = require('url');
const net = require('net');
const _ = require('lodash');
const dayjs = require('dayjs');
const isPortReachable = require('is-port-reachable');
const randomstring = require('randomstring');
const simpleEncryptor = require('simple-encryptor');
const BindSocketsReverse = require('./bindSockets_reverse');
const { EventEmitter } = require('events');
const logger = require('debug')('WstServerReverse');
let connected = {}
class WstServerReverse {
	constructor(secret, httpServer) {
		const wsServerForControl = new WebSocketServer({
			httpServer,
			useNativeKeepalive: true,
			keepalive: true,
			autoAcceptConnections: false,
		});

		this.latestConnectedTime = dayjs().unix();
		this.secret = secret;
		this.httpServer = httpServer;
		this.wsServerForControl = wsServerForControl;
		this.newWSTCPEmitter = new EventEmitter();
	}

	start({ minRandomPort = 1, maxRandomPort = 16 ** 4 - 1, path }) {
		this.wsServerForControl.on(
			'request',
			((_this) => {
				return async (request) => {
					// One TCP Server for each client WS Request
					logger('wsServerForControl request');

					// parse to http uri
					const uri = new URLSearchParams(_.get(request, 'httpRequest.url', '').replace(path, '').replace(/^\//, ''));
					const dst = uri.get('dst');
					const namespace = uri.get('namespace')
				//	console.log('request', _.get(request, 'httpRequest.url', ''))


					if (dst && path === uri.get('path')) {
						if(connected[namespace])return 
						const fetchPort = async () => {
							const getPort = () => {
								const portNumber = _.random(minRandomPort, maxRandomPort);
								return isPortReachable(portNumber, {
									host: 'localhost',
								}).then((reachable) => (reachable ? getPort() : portNumber));
							};

							return dst === 'random' ? getPort() : dst;
						};

						const portTcp = await fetchPort();
						request.tcpServer = net.createServer();
						// start tcp server
						request.tcpServer.listen(portTcp);
						logger(`Created TCP server on port: ${portTcp}`);
						this.newWSTCPEmitter.emit('inited', { port: portTcp, namespace })
						// add accept header
						request.wsConnectionForControl = request.accept('tunnel-protocol', request.origin);

						// start message
						logger('WS Connection for Control Created');
						connected[namespace]=true
						request.wsConnectionForControl.on('error', logger);
						request.wsConnectionForControl.on('close', (reasonCode, description) => {
							logger(
								`WebSocket Control Peer ${_.get(
									request,
									'wsConnectionForControl.remoteAddress'
								)} disconnected for: ${description}`
							);
							logger(`Close TCP server on port ${portTcp}`);

							request.tcpServer.close();
							this.newWSTCPEmitter.emit('close', { port: portTcp, namespace })
							delete connected[namespace];
						});
					} else {
						logger('Request for Data WS Socket');

						// emit request
						this.newWSTCPEmitter.emit('created', request);
					}
					if (!request.tcpServer) return
					// add error callback
					request.tcpServer.on('error', logger);
					let n = 0
					// Manage TCP Connection
					request.tcpServer.on(
						'connection',
						((_this) => {
							return (tcp) => {
								tcp.on('error', () => {
									console.log('tcp error')
								})
							
								logger('tcpServer connection');
								n++
								//		if(n>5)return 
								// Putting in pause the tcp connection waiting the new socket WS Socket for data
								tcp.pause();

								// create connection
								const idConnection = randomstring.generate(20);
								const msgForNewConnection = `NC:${idConnection}`;

								// send utf message for new connection
								request.wsConnectionForControl.sendUTF(msgForNewConnection);
								this.newWSTCPEmitter.once(
									'created',
									((_this) => {
										return (request) => {
											console.log('created')
											try {
												const encryptor = simpleEncryptor(_this.secret);
												const requestUrl = _.get(request, 'httpRequest.url', '');
												const uri = new URLSearchParams(requestUrl.replace(path, '').replace(/^\//, ''));
												const id = uri.get('id');
												const token = uri.get('token');
												const time = _.toInteger(encryptor.decrypt(token));

												// check connection state
												if (idConnection === id && dayjs().unix() - time < 30) {
													const ws = request.accept('tunnel-protocol', request.origin);
													const bind = new BindSocketsReverse(ws, tcp);

													// start bind service
													bind.start();

													// Resuming of the tcp connection after WS Socket is just created
													tcp.resume();

													// mark this object as connected
													_this.isConnected = true;

													// save latest connected time
													_this.latestConnectedTime = dayjs().unix();
												}
											} catch (error) {
												logger(error);
											}
										};
									})(this)
								);
							};
						})(_this)
					);
				};
			})(this)
		);
	}
}

module.exports = WstServerReverse;
